"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var fs_1 = require("fs");
var EventEmitter = require("events");
/*--- useful functions ---*/
var hasOwnProperty = function (target, name) { return Object.prototype.hasOwnProperty.call(target, name); };
var defaultToObject = {
    get: function (target, name) {
        return !hasOwnProperty(target, name)
            ? {}
            : target[name];
    }
};
var decode = function (buf, encoding) {
    if (encoding === 'utf16') {
        if (buf[0] === 0xFE && buf[1] === 0xFF) {
            encoding = 'utf16be';
            buf = buf.slice(2);
        }
        else if (buf[1] === 0xFE && buf[0] === 0xFF) {
            encoding = 'utf16le';
            buf = buf.slice(2);
        }
        else {
            encoding = 'utf16be';
        }
    }
    switch (encoding) {
        case 'utf16be':
            if (buf.length === 0)
                return '';
            var res = Buffer.alloc(buf.length);
            var i = 0;
            for (; i < buf.length - 1; i += 2) {
                res[i] = buf[i + 1];
                res[i + 1] = buf[i];
            }
            return res.toString('utf16le', 0, i);
        default:
            return buf.toString(encoding);
    }
};
var unicodeTerminator = Buffer.from([0, 0]);
var asciiTerminator = Buffer.from([0]);
var parseFrameEncoding = function (frame) {
    return frame[0] === 0 ? { encoding: 'ascii', terminator: asciiTerminator }
        : frame[0] === 1 ? { encoding: 'utf16', terminator: unicodeTerminator }
            : frame[0] === 2 ? { encoding: 'utf16be', terminator: unicodeTerminator }
                : frame[0] === 3 ? { encoding: 'utf8', terminator: asciiTerminator }
                    : { encoding: 'ascii', terminator: asciiTerminator };
};
var syncSafeIntToInt = function (x) { return (x & 127) + ((x & 32512) >>> 1); };
var decodeFrame = function (id, frame) {
    if (id === 'TXXX') {
        var _a = parseFrameEncoding(frame), encoding = _a.encoding, terminator = _a.terminator;
        var descriptorEndIndex = frame.indexOf(terminator, 1) + terminator.length;
        var descriptor = decode(frame.slice(1, descriptorEndIndex - terminator.length), encoding);
        var text = decode(frame.slice(descriptorEndIndex), encoding);
        return { id: id, descriptor: descriptor, text: text };
    }
    else if (id.startsWith('T')) {
        var encoding = parseFrameEncoding(frame).encoding;
        var text = decode(frame.slice(1), encoding);
        return { id: id, text: text };
    }
    else if (id === 'WXXX') {
        var _b = parseFrameEncoding(frame), encoding = _b.encoding, terminator = _b.terminator;
        var descriptorEndIndex = frame.indexOf(terminator, 1) + terminator.length;
        var descriptor = decode(frame.slice(1, descriptorEndIndex - terminator.length), encoding);
        var url = decode(frame.slice(descriptorEndIndex), 'ascii');
        return { id: id, descriptor: descriptor, url: url };
    }
    else if (id.startsWith('W')) {
        var url = decode(frame, 'ascii');
        return { id: id, url: url };
    }
    else if (id === 'USLT') {
        var _c = parseFrameEncoding(frame), encoding = _c.encoding, terminator = _c.terminator;
        var languageCode = decode(frame.slice(1, 4), 'ascii');
        var descriptorEndIndex = frame.indexOf(terminator, 4) + terminator.length;
        var descriptor = decode(frame.slice(4, descriptorEndIndex - terminator.length), encoding);
        var text = decode(frame.slice(descriptorEndIndex), encoding);
        return { id: id, languageCode: languageCode, descriptor: descriptor, text: text };
    }
    else if (id === 'SYLT') {
        var _d = parseFrameEncoding(frame), encoding = _d.encoding, terminator = _d.terminator;
        var languageCode = frame.toString('ascii', 1, 4);
        var timeStampFormat = frame.toString('hex', 4, 5);
        var lengthOfBytes = timeStampFormat === '01' || timeStampFormat === '02' ? 4 : 0;
        var type = frame.toString('hex', 5, 6);
        var descriptorEndIndex = frame.indexOf(terminator, 6) + terminator.length;
        var descriptor = decode(frame.slice(6, descriptorEndIndex - terminator.length), encoding);
        var syncText = [];
        var unprocessedLyrics = frame.slice(descriptorEndIndex);
        while (unprocessedLyrics.length) {
            var textEnd = unprocessedLyrics.indexOf(terminator) + terminator.length;
            var text = decode(unprocessedLyrics.slice(0, textEnd - terminator.length), encoding);
            var timeStampBytes = unprocessedLyrics.slice(textEnd, textEnd + lengthOfBytes);
            var timeStamp = timeStampBytes.readUIntBE(0, timeStampBytes.length);
            syncText.push([timeStamp, text]);
            unprocessedLyrics = unprocessedLyrics.slice(textEnd + lengthOfBytes);
        }
        return { id: id, syncText: syncText, type: type, descriptor: descriptor, languageCode: languageCode, timeStampFormat: timeStampFormat };
    }
    else if (id === 'APIC') {
        var _e = parseFrameEncoding(frame), encoding = _e.encoding, terminator = _e.terminator;
        var mimeTypeEndIndex = frame.indexOf(terminator, 1);
        var mimeType = decode(frame.slice(1, mimeTypeEndIndex), encoding);
        var descriptorEndIndex = mimeTypeEndIndex + terminator.length + 1;
        var descriptor = frame.slice(mimeTypeEndIndex + terminator.length, descriptorEndIndex).toString('hex');
        var descriptionEndIndex = frame.indexOf(terminator, descriptorEndIndex);
        var description = decode(frame.slice(descriptorEndIndex, descriptionEndIndex), encoding);
        var buffer = frame.slice(descriptionEndIndex + 1);
        return { id: id, mimeType: mimeType, descriptor: descriptor, description: description, buffer: buffer };
    }
    else {
        return { id: id, buffer: frame };
    }
};
function readId3(input, needed) {
    var parser = new EventEmitter();
    var stream = typeof input === 'string'
        ? fs_1.createReadStream(input)
        : input;
    // pre-stage: defining things which will be needed in state 0
    var headerSize = 10;
    var frameHeaderSize = 10;
    var state = 0;
    var end = false;
    var header = Buffer.alloc(0);
    parser.on('end', function () {
        end = true;
        if (typeof input === 'string') {
            stream.close();
        }
    });
    // state 0: parsing header
    var majorVersion;
    var patchVersion;
    var framesSize;
    var frameHeader = Buffer.alloc(0);
    // state 1: parsing frame header
    var frameName;
    var frameValue = Buffer.alloc(0);
    var frameValueLength;
    // state 2: parsing frame content
    /* after parsed, it goes back to state 1 so no need to define new variables */
    stream.on('data', function (chunk) {
        if (end)
            return;
        // state 0: parsing header
        if (state === 0) {
            header = Buffer.concat([header, chunk]);
            if (header.length < headerSize)
                return;
            if (header.toString('ascii', 0, 3) !== 'ID3') {
                parser.emit('error', new Error('There is no ID3 tag!'));
            }
            majorVersion = header[3];
            patchVersion = header[4];
            if (majorVersion !== 3 && majorVersion !== 4) {
                parser.emit('error', new Error('ID3v2.' + majorVersion + ' is not supported!'));
            }
            framesSize = (header[6] << 21) + (header[7] << 14) + (header[8] << 7) + header[9];
            chunk = header.slice(headerSize);
            header = undefined;
            state = 1;
        }
        framesSize -= chunk.length;
        if (framesSize < 0) {
            chunk = chunk.slice(0, chunk.length + framesSize);
            framesSize = 0;
        }
        while (true) {
            // state 1: parsing frame headers
            if (state === 1) {
                frameHeader = Buffer.concat([frameHeader, chunk]);
                if (frameHeader.length < frameHeaderSize)
                    break;
                var frameHeaderStartIndex = frameHeader.toString('ascii').search(/[A-Z]{3}[A-Z0-9]/);
                if (frameHeaderStartIndex === -1) {
                    frameHeader = frameHeader.slice(frameHeader.length - 3);
                    break;
                }
                var trueFrameHeader = frameHeader.slice(frameHeaderStartIndex);
                frameName = trueFrameHeader.toString('ascii', 0, 4);
                frameValueLength =
                    majorVersion === 3
                        ? parseInt(trueFrameHeader.toString('hex', 4, 8), 16)
                        : syncSafeIntToInt(trueFrameHeader.slice(4, 8).readUIntBE(0, 4));
                chunk = frameHeader.slice(frameHeaderSize);
                frameHeader = Buffer.alloc(0);
                state = 2;
            }
            // state 2: parsing frame values
            if (state === 2) {
                frameValue = Buffer.concat([frameValue, chunk]);
                if (frameValue.length < frameValueLength)
                    break;
                var formattedFrame = decodeFrame(frameName, frameValue.slice(0, frameValueLength));
                parser.emit('frame', formattedFrame);
                if (end)
                    return;
                chunk = frameValue.slice(frameValueLength);
                frameValue = Buffer.alloc(0);
                state = 1;
            }
        }
        if (framesSize === 0) {
            parser.emit('end');
        }
    });
    if (Array.isArray(needed)) {
        return new Promise(function (resolve, reject) {
            var leftNeeded = new Set(needed);
            var result = new Map();
            parser.on('frame', function (x) {
                if (leftNeeded.delete(x.id)) {
                    result.set(x.id, x);
                }
                if (typeof x.type === 'string') {
                    var actualId = x.id + ':' + x.type;
                    if (leftNeeded.delete(actualId)) {
                        result.set(actualId, x);
                    }
                    if (typeof x.descriptor === 'string') {
                        var actualId2 = actualId + ':' + x.descriptor;
                        if (leftNeeded.delete(actualId2)) {
                            result.set(actualId2, x);
                        }
                    }
                }
                else if (typeof x.descriptor === 'string') {
                    var actualId = x.id + ':' + x.descriptor;
                    if (leftNeeded.delete(actualId)) {
                        result.set(actualId, x);
                    }
                }
                if (leftNeeded.size === 0) {
                    parser.emit('end');
                }
            });
            parser.on('end', function () {
                resolve(result);
            });
        });
    }
    else if (typeof needed === 'undefined') {
        return new Promise(function (resolve, reject) {
            var result = new Map();
            parser.on('frame', function (x) {
                result.set(x.id, x);
                if (typeof x.descriptor === 'string') {
                    var actualId = x.id + ':' + x.descriptor;
                    result.set(actualId, x);
                }
            });
            parser.on('end', function () {
                resolve(result);
            });
        });
    }
    else {
        return parser;
    }
}
exports.readId3 = readId3;
;
